/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"; you may not use
 * this file except in compliance with the License. You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.statement;

import io.confluent.ksql.parser.tree.Statement;

/**
 * Injects a {@link ConfiguredStatement} with more information such as
 * schema or kafka topic.
 */
public interface Injector {

  /**
   * @param statement the statement to inject information
   * @param <T>       the type of the statement
   *
   * @return a statement with information injected
   * @implNote this method should return a new instance if anything changed
   */
  <T extends Statement> ConfiguredStatement<T> inject(ConfiguredStatement<T> statement);

}
